#!/usr/bin/python
import math, time, random, sys

'''
# Dieses Programm berechnet die Determinanten von n x n Matrizen (n=1, ....,10),
# die vorab mit Zufallszahlen zwischen 0 und 1 besetzt werden, mit dem
# Entwicklungssatz. Die Rechenzeiten fuer die Berechnung der Determinanten
# wird mit ausgegeben.
# Dies zeigt, wie dramatisch ein Rechenaufwand, der proportional zu n! ist,
# anwaechst.
# Will man Determinanten noch groesserer Matrizen berechnen, reicht es im
# untenstehenden Befehl
#         for n in range(1,11):
# die Zahlen 1 und 11 zu aendern. Die 1 ist durch die erste
# Zahl zu ersetzen, die als kleinste Dimension einer Matrix gewaehlt werden
# soll, die Zahl anstelle von 11 muss um 1 groesser gewaehlt werden als die
# Zahl, die als groesste Dimension gewaehlt werden soll.
'''

def calcasub(a,asub,n,i):
# Nach dem Entwicklungssatz wird die Berechnung der Determinante
# einer Matrix n x n Matrix a auf die Berechnung von n Determinanten
# von (n-1) x (n-1) Untermatrizen zurueckgefuehrt. Die Funktion
# calcasub berechnet die naechste Untermatrix und speichert sie in asub.
   for k in range(1,n):
      for j in range(0,i):
         asub[k-1][j]=a[k][j]
      for j in range(i+1,n):
         asub[k-1][j-1]=a[k][j]
   return asub
         
def detcalc(a,n):
# Diese Funktion berechnet die Determinante der n x n Matrix a mit dem
# Entwicklungssatz. Realisiert wird dies durch rekursive Funktionsaufrufe
# der Funktion detcalc.
   if n >=3:
      det = 0.
      asub=range(n-1)
      for i in range(n-1):
         asub[i]=range(n-1)
      for i in range(n):
         # Nach dem Entwicklungssatz wird die Berechnung der Determinante
         # einer Matrix n x n Matrix a auf die Berechnung von n Determinanten
         # von (n-1) x (n-1) Untermatrizen zurueckgefuehrt. Die Funktion
         # calcasub berechnet die naechste Untermatrix, die in asub gespeichert
         # wird.
         asub = calcasub(a,asub,n,i)
         # Die naechste Unterdeterminante muss nur berechnet werden, wenn
         # a[0][i] ungleich 0 ist.
         if a[0][i]!=0.:
            det = det + (-1.)**i*a[0][i]*detcalc(asub,n-1)
      return det
   elif n==2:
      det = a[0][0]*a[1][1]-a[1][0]*a[0][1]
      return det
   elif n==1:
      return a[0][0]
   else:
      print "negative entry for matrix dimension; program stops"
      return

#################
# HAUPTPROGRAMM #
#################

print 
print
print "Dieses Programm berechnet die Determinanten von n x n Matrizen "
print "fuer n=1, ....,10 mit dem Entwicklungssatz."
print "Dabei werden die Matrizen mit Zufallszahlen zwischen 0 und 1 besetzt."
print "Die Rechenzeiten wachsen proportional zu n! und werden mit dem jeweiligen "
print "Wert der Determinante ausgegeben."
print
print 
print "   n     time         det "
print " ----------------------------"
for n in range(1,11):
   a=range(n)
   for i in range(n):
      a[i]=range(n)  
      for j in range(n):
         a[i][j]=random.random()
   time1=time.time()
   determa=detcalc(a,n)
   time2=time.time()
   print "%4i  %8.3f  %12.6f" %(n, time2-time1, determa)
sys.stdin.readline()

