#!/usr/bin/python

# full multigrid program for the solution of the 2D boundary value problem
#        delta u = f 
#              u = g
# in the unit square

import math, time, sys

def frhs(x,y):
   f = 12.*x**2+24.*y**2
#   f = -25.*math.pi**2*math.sin(3.*math.pi*x)* math.sin(4.*math.pi*y)
   return f
def urhs(x,y):
   u = x**4+2.*y**4 
#   u = math.sin(3.*math.pi*x)* math.sin(4.*math.pi*y) 
   return u
def intadd(u,uc,nc):
   for ic in range(1,nc):
      i = 2 * ic
      for jc in range(1,nc):
         j = 2 * jc 
         u[i][j]=u[i][j]+uc[ic][jc]
         u[i-1][j]= u[i-1][j]+0.5*(uc[ic][jc]+uc[ic-1][jc])
         u[i][j-1]= u[i][j-1]+0.5*(uc[ic][jc]+uc[ic][jc-1])
         u[i-1][j-1] = u[i-1][j-1]\
                      +0.25*(uc[ic][jc]+uc[ic][jc-1]+uc[ic-1][jc]+uc[ic-1][jc-1])
   return u
def intpol4(u,uc,nc):
#--------------------------------------------------
# 4th order interpolation (3rd order if nc==2) 
# Here, nc is the number of intervals on the coarse grid
#--------------------------------------------------

# interpolation of fine grid points which belong also to the coarse grid
   n = 2*nc
   for ic in range(0,nc+1):
      i = 2*ic
      for jc in range(0,nc+1):
         j = 2*jc
         u[i][j] = uc[ic][jc]
#------------------------
# 3rd order interpolation if not enough coarse grid points for 4th order
# interpolation available
#------------------------
   if nc == 2:
      z18 = .125
      z38 = .375
      z68 = .75
# interpolation of remaining points in coarse grid columns
      for ic in range(0,nc+1):
         i = 2*ic
         u[i][1] = z38*uc[ic][0] + z68*uc[ic][1] - z18*uc[ic][2]
         u[i][3] = z38*uc[ic][2] + z68*uc[ic][1] - z18*uc[ic][0]
# interpolation of remaining points in fine grid rows
      for j in range(0,n+1):
         u[1][j] = z38*u[0][j] + z68*u[2][j] - z18*u[4][j]
         u[3][j] = z38*u[4][j] + z68*u[2][j] - z18*u[0][j]
#------------------------
# 4th order interpolation 
#------------------------
   else:
      z116=1./16. 
      z516=5./16. 
      z916=9./16. 
      z1516=15./16. 
# interpolation of remaining points in coarse grid columns
      for ic in range(0,nc+1):
         i = 2*ic
         u[i][1] = z516*(uc[ic][0]-uc[ic][2]) + z1516*uc[ic][1] + z116*uc[ic][3]
         u[i][n-1] = z516*(uc[ic][nc]-uc[ic][nc-2]) + z1516*uc[ic][nc-1] + z116*uc[ic][nc-3]
         for jc in range(1,nc-1):
            j = 2*jc + 1
            u[i][j] = z916*(uc[ic][jc]+uc[ic][jc+1])-z116*(uc[ic][jc-1]+uc[ic][jc+2])
#interpolation of remaining points in fine grid rows
      for j in range(0,n+1):
         u[1][j] = z516*(u[0][j]-u[4][j]) + z1516*u[2][j] + z116*u[6][j]
         u[n-1][j] = z516*(u[n][j]-u[n-4][j]) + z1516*u[n-2][j] + z116*u[n-6][j]
         for i in range(3,n-2,2):
            u[i][j]=z916*(u[i-1][j]+u[i+1][j])-z116*(u[i-3][j]+u[i+3][j])
   return u
def half_injection(u,f,n,nc,h2):
   fc = range(nc)
   for i in range (0,nc):
      fc[i]= f[2*i][0:n:2]
   for ic in range (1,nc-1):
      i = 2 * ic
      for jc in range (1,nc-1):
         j = 2 * jc
         fc[ic][jc] = 0.5*(fc[ic][jc]\
                     +(u[i-1][j]+u[i][j-1]+u[i+1][j]+u[i][j+1]-4.0*u[i][j])/h2)
   return fc
def cycle(u,f,h2,n):
   u = gsiteration(u,f,h2,n)

   nc = (n+1)/2
   hc = 1./(nc-1)
   h2c = hc*hc
   fc = half_injection(u,f,n,nc,h2)
   uc=range(nc)
   for i in range (0,nc): 
      uc[i]=range(nc)
   for i in range (0,nc):
      for j in range (0,nc):
         uc[i][j]=0.
   if n <= 5:
      uc = gsiteration(uc,fc,h2c,nc)
   else:
      uc=cycle(uc,fc,h2c,nc)
   u = intadd(u,uc,nc)
   resi=resicalc(u,f,h2,n)
   u = gsiteration(u,f,h2,n)
   u = gsiteration(u,f,h2,n)
   return u
def gsiteration(u,f,h2,n):
   for i in range (1,n-1,2):
      for j in range (1,n-1,2):
         u[i][j]=0.25*(f[i][j]*h2+u[i-1][j]+u[i][j-1]+u[i+1][j]+u[i][j+1])
   for i in range (2,n-1,2):
      for j in range (2,n-1,2):
         u[i][j]=0.25*(f[i][j]*h2+u[i-1][j]+u[i][j-1]+u[i+1][j]+u[i][j+1])
   for i in range (1,n-1,2):
      for j in range (2,n-1,2):
         u[i][j]=0.25*(f[i][j]*h2+u[i-1][j]+u[i][j-1]+u[i+1][j]+u[i][j+1])
   for i in range (2,n-1,2):
      for j in range (1,n-1,2):
         u[i][j]=0.25*(f[i][j]*h2+u[i-1][j]+u[i][j-1]+u[i+1][j]+u[i][j+1])
   return u
def resicalc(u,f,h2,n):
   resi=0.
   for i in range (1,n-1):
      for j in range (1,n-1):
         resi = max(resi,abs(f[i][j]\
               +(u[i-1][j]+u[i][j-1]+u[i+1][j]+u[i][j+1]-4.0*u[i][j])/h2))
   return resi
def maxabs(u,f,h,n):
   maxerr=0.
   for i in range (1,n-1):
      for j in range (1,n-1):
         maxerr = max(maxerr,abs(u[i][j]-urhs(i*h,j*h)))
   return maxerr

###################################
# main program for full multigrid #
###################################

print
print 'Full-Multigrid-Programm zur Loesung der zweidimensionalen Poissonschen Randwertaufgabe'
#print 'Anzahl der Mehrgitteriterationen pro Level:  1'
levels=10
print 'Anzahl der Gitter: ', levels
time1=time.time()
for lev in range(1,levels+1):
   n=2**lev+1
   print
   print 'Gitter : ', lev
   print 'Intervalle in jeder Richtung: ',n-1 
   print 'Gitterpunkte und Unbekannte: ',n*n
   h=1./(n-1)
   h2=h*h
   x=[]
   y=[]
   for i in range (0,n):
      x.append(i*h)
      y.append(i*h)
   f1=range(n)
   f=range(n)
   u=range(n)
   for i in range (0,n):
      f[i]=f1[:]
      u[i]=f1[:]
   if lev == 1:
      for i in range (1,n-1):
         for j in range (1,n-1):
            u[i][j]=0.
   else:
      u = intpol4(u,uc,(n-1)/2)
   for i in range (0,n):
      u[i][0]=urhs(x[i],0.)
      u[i][n-1]=urhs(x[i],1.)
      u[0][i]=urhs(0.,y[i])
      u[n-1][i]=urhs(1.,y[i])
   for i in range (1,n-1):
      for j in range (1,n-1):
         f[i][j]=-frhs(x[i],y[j])
   uc = cycle(u,f,h2,n)
   print 'Auf Gitter ',lev,' erreichter maximaler absoluter Fehler: ',\
         maxabs(uc,f,h,n)
   time2=time.time()
   print "Benoetigte Zeit zur Loesung des GLS mit ",n*n," Unbekannten: %.2f sec"\
         %(time2-time1)
sys.stdin.readline()
